from pathlib import Path
from typing import Dict
from google.cloud import storage
import os
import markdown
import time
import sys
import bloggen.generate as generate

class Site:
    def __init__(self, user_config):
        self.user_config = user_config
        self.client = storage.Client()
        self.bucket_name = self.user_config.active_config['data']['buckets'][0]
        self.host_url = 'https://storage.cloud.google.com/'
        self.bucket = self.retrieve_bucket(self.bucket_name)
        # default setting all buckets to public
        self.set_privacy(public=True)
        # below line assumes that the user generates a static site in their cwd
        self.static_site_root = Path(os.getcwd())

    def add(self, path_to_md):
        with open(path_to_md) as f:
            md_input = f.read()
        filename = os.path.basename(path_to_md)[:-3]
        outfile = f'notes/{filename}.html'
        html = markdown.markdown(md_input)
        with open(outfile, 'w') as f:
            f.write(html)
        blob = self.bucket.blob(outfile)
        blob.upload_from_filename(outfile)
                                            # if no param is passed, assume that the user wants to publish the current dir of markdown files as a static-site
    def publish(self, path_to_static_site:Path): # path_to_static_site:str=generate.get_static_site_dir()
        self.static_site_root = path_to_static_site
        # need to prepare for hosting site:
            # replace all local hrefs with refs to files on cloud.
            # Will follow this pattern: https://storage.cloud.google.com/first-bloggen-bucket/static-site/notes/test%20copy%202.html
        notes_root = f"{self.host_url}{self.bucket_name}/static-site/notes/"
        generate.prep_for_hosting(self.static_site_root, notes_root)

        # learn which bucket to use
        target_bucket_name = self.user_config.active_config['data']['buckets'][0] if len(self.user_config.active_config['data']['buckets']) < 2 else input(f"Which bucket? {self.user_config.active_config['data']['buckets']}")
        # learn if bucket exists
        target_bucket = self.retrieve_bucket(target_bucket_name)

        if target_bucket:
            print('uploading bucket')
            self.upload_site(path_to_static_site.as_posix())
            bucket_url = f"{self.host_url}{target_bucket_name}/static-site/index.html"
            print(f'Your notes are available as html at {bucket_url}')
        else:
            print("Bucket not insantiated",file=sys.stderr)

    def generate(self, path_to_md_dir: Path, site_info: Dict):
        self.static_site_root = Path.joinpath(path_to_md_dir.parent,'static-site')
        generate.static_site(self.static_site_root, site_info)
        generate.html_files(path_to_md_dir, self.static_site_root)
        generate.index_html(self.static_site_root)
        print(f'See your local site at {os.path.join(self.static_site_root,"index.html")}')

    def get_bucket(self, name):
        return self.client.get_bucket(name)

    def make_bucket(self, name):
        self.bucket: storage.Bucket = self.client.create_bucket(name)

    def set_privacy(self, public: bool):
        if public:
            self.bucket.make_public
        else:
            self.bucket.make_private

    def upload_site(self, path_to_dir:str):
        root = os.path.basename(path_to_dir)
        for path, _, files in os.walk(path_to_dir):
            for name in files:
                print(f'uploading file {name}')
                root_index = path_to_dir.find(root)
                blob_path = os.path.join(path, name).replace('\\','/')[root_index:]
                blob = self.bucket.blob(blob_path)
                local_path = os.path.join(path, name)
                blob.upload_from_filename(local_path)
                blob.make_public()

    def upload_files(self, path_to_dir) -> str:
        for path, _, files in os.walk(path_to_dir):
            for name in files:
                print(f'uploading {name}')
                path_local = os.path.join(path, name)
                blob_path = path_local.replace('\\','/')
                blob = self.bucket.blob(blob_path)
                blob.upload_from_filename(path_local)

    def retrieve_bucket(self, bucket_name):
        try:
            return self.get_bucket(bucket_name)
        except:
            bucket = self.make_bucket(bucket_name)
            print(f'Creating a new bucket named {bucket_name}')
            time.sleep(4)
            return bucket



import os
import json
from bloggen.providers import config_functions as configure_by_provider
from pathlib import Path, PurePath

class Configure:
    
    def __init__(self):
        self.config_path = PurePath.joinpath(Path(__file__).parent, '.bloggen/config.json')
        self.active_config = self.get_active_config()
        if not self.active_config:
            print("No config selected. Using default.")
            print("Create a new configuration with --config option. Activate a configuration with --config [name of config here]")
            self.set_active_config('default')
        self.apply_config(self.active_config)
        print( f'Active config: {self.active_config["name"]}' )
        
    def init_bloggen(self):
        active_config = self.active_config
        backend_provider = active_config['data']['backend']
        provider_config = self.get_provider_config(backend_provider)
        configure_by_provider[backend_provider](provider_config)

    def user_config_exists(self, name):
        return name in self.list_config_names()

    def list_config_names(self):
        with open(self.config_path) as f:
            configs = json.load(f)
        config_names = [config['name'] for config in configs]
        return config_names

    def get_active_config(self):
        with open(self.config_path) as f:
            configs = json.load(f)
        return list(filter(lambda config: config['active'] == True, configs))[0]

    def set_active_config(self, name):
        """
        Change the active key to true in the config.json file for the given name.
        """
        with open(self.config_path) as f:
            configs: list = json.load(f)
        for user_config in configs:
            if user_config['name'] == name:
                if self.valid_user_config(user_config):
                    self.apply_config(user_config)
                user_config['active'] = True
            else:
                user_config['active'] = False
        with open(self.config_path, 'w') as f:
            json.dump(configs, f)
        self.active_config = user_config['name']

    def get_provider_config(self, backend_provider):
        with open('.bloggen/provider.json') as f:
            providers = json.load(f)
        return providers[backend_provider]

    def add_user_config(self, new_config):
        with open(self.config_path) as f:
            config = json.load(f)
        config.append(new_config)
        with open(self.config_path, 'w') as f:
            json.dump(config, f)
    
    def update_user_config(self, key, value):
        updated_config = self.active_config
        updated_config['data'][key] = value
        with open(self.config_path) as f:
            config = json.load(f)
        for i, user_config in enumerate(config):
            if user_config['active'] == True:
                config[i] = updated_config
        with open(self.config_path, 'w') as f:
            json.dump(config, f)
        print(f'Set {key} of {self.active_config["name"]} to {value}')
        self.apply_config(self.active_config)

    def create_user_config(self):
        print("creating config")
        new_config = {
            'name': '',
            'active': False,
            'data': {
                'backend': 'gcp',
                "credentials": "path_to_credentials",
                "buckets": []
            }
        }
        new_config['name'] = input("Name: ")
        new_config['data']['credentials'] = input("Path to your GCP credentials json: ")
        new_config['data']['buckets'].append(input("Name of GCP Bucket: ").lower().replace(' ', '-'))
        self.add_user_config(new_config)


    def add_bucket(self):
        pass

    def valid_user_config(self, user_config):
        if not user_config['data']['credentials']:
            print("Please set the path to your credential file with this command: bloggen --config credential=path/to/credential")
            return False
        else:
            config_path = Path(user_config['data']['credentials'])
            if not Path.exists(config_path):
                print(f"Could not find a file at {user_config['data']['credentials']}")
                print("Please set the path to your credential file with this command: bloggen --config credential=path/to/credential")
                return False
            return True

    def apply_config(self, config):
        # set credentials of blog host 
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = config['data']['credentials']
        os.environ['GCLOUD_PROJECT'] = config['data']['project']


import os

def config_gcp(gcp_config):
    if not os.getenv(gcp_config['credential_var']):
        print('GCP credential not found in this environment. Could not find variable GOOGLE_APPLICATION_CREDENTIALS')
        print('Step 1: Get a service token from GCP. Follow this guide https://cloud.google.com/storage/docs/reference/libraries#setting_up_authentication')
        print('Step 2: Save the JSON to your machine.')
        print('Step 3: Set environment variable GOOGLE_APPLICATION_CREDENTIALS to path of your JSON.')

config_functions = {
    'gcp': config_gcp
}



import json
import os
from pyclbr import Function
from sys import stderr, exit
from typing import Dict
import markdown
import bs4
from pathlib import Path
import shutil
import pathlib

    # gets the dir of static-site if it is the same dir as bloggen itself.
def get_static_site_dir():
    dir = os.path.dirname(__file__)
    path = Path(dir)
    path = path.parent.absolute()
    path = path.joinpath('static-site')
    return path.absolute().as_posix()

def html_files(input_dir:Path, output_dir:Path):
    not_md = []
    for path in Path.iterdir(input_dir):
        filename = path.name
        if filename.endswith('.md'):
            print('file: '+ filename)
            __html_file(Path.joinpath(input_dir,filename), Path.joinpath(output_dir, 'notes'))
        else:
            not_md.append(filename)
    if not_md:
        for file in not_md:
            print(f'{file} is not a markdown file')

def __html_file(input_file:str,output_dir:Path):
    """
    Converts a single markdown file to html and writes it to a given dir.
    """
    filename = os.path.basename(input_file).replace('.md','.html')
    outfile = Path.joinpath(output_dir,filename)
    with open(input_file) as f:
            md_input = f.read()
    html = markdown.markdown(md_input)
    with open(outfile, 'w') as f:
        f.write(html)

def index_html(path_to_site: Path):
    cook_soup(path_to_site, recipe=index_md_files)

def prep_for_hosting(path_to_static_site:Path, notes_root:str, ):
    # access config file
    # switch_index_references(notes_root)
    cook_soup(path_to_static_site, prep_for_cloud, ingredients={'notes_root': notes_root})

def cook_soup(path_to_site: Path, recipe:Function, ingredients:Dict={}):
    path_to_posts_dir = Path.joinpath(path_to_site,'notes')
    path_to_index = Path.joinpath(path_to_site,'index.html')
    with open(path_to_index) as in_f:
        txt = in_f.read()
        soup = bs4.BeautifulSoup(txt, features="html.parser")
    post_names:list[str] = sorted([note.name for note in Path.iterdir(path_to_posts_dir)])
    for post_name in post_names:
        ingredients['post_name'] = post_name
        recipe(ingredients, soup)
    with open(path_to_index, "w") as out_f:
        out_f.write(str(soup))  

def index_md_files(data:dict, soup):
    """
    Scans generated html files and creates links to them on the index.html page.
    For every html in ../notes dir of static site, add link to index
    This function assumes that there are notes in the ../static-site/notes dir
    """
    post_name = data['post_name']
    new_li = soup.new_tag('li')
    new_anchor = soup.new_tag('a', href=os.path.join('notes',post_name))
    post_name = post_name.removesuffix('.html')
    new_anchor.attrs['id'] = post_name.replace(' ','')
    new_anchor.string = post_name
    new_li.append(new_anchor)
    soup.body.ul.append(new_li)

def prep_for_cloud(data:dict, soup):
    post_name = data['post_name']
    notes_root = data['notes_root']
    tag_id = post_name.removesuffix('.html').replace(' ','')
    anchor:bs4.Tag = soup.find(id=tag_id)
    if anchor:
        anchor.attrs['href'] = notes_root + post_name.replace(' ','%20')
    else:
        print(f"Index.html linking won't work. Could not find a tag with tag_id '{tag_id}'. As a result, I could not update the link.",stderr)

def switch_index_references(notes_root:str,path_to_site:str=get_static_site_dir()):
    path_to_posts_dir = Path.joinpath(path_to_site,'notes')
    path_to_index = Path.joinpath(path_to_site,'index.html')
    with open(path_to_index) as in_f:
        txt = in_f.read()
        soup = bs4.BeautifulSoup(txt, features="html.parser")
    post_names:list[str] = sorted([note.name for note in Path.iterdir(path_to_posts_dir)])
    for post_name in post_names:
        tag_id = post_name.removesuffix('.html').replace(' ','')
        anchor:bs4.Tag = soup.find(id=tag_id)
        if anchor:
            anchor.attrs['href'] = notes_root + post_name.replace(' ','%20')
        else:
            print(f"Index.html linking won't work. Could not find a tag with tag_id '{tag_id}'. As a result, I could not update the link.",stderr)

    with open(path_to_index, "w") as out_f:
        out_f.write(str(soup))    

def static_site(static_site_root: Path, site_info: Dict):
    """
    1. Generate static site directory structure
        1a. create index.html file
        1b. Create open html tags
    2. for very md file in folder, convert to html
        2a. write to a file in notes dir
        2b. add link of each file to open html tag in index.html
    3. 
    """
    if Path.exists(static_site_root):
        print(f'static-site dir already exists at ${static_site_root}. Canceling generation.')
        exit()
        
    os.mkdir(static_site_root)
    os.mkdir(Path.joinpath(static_site_root, 'notes'))
    os.mkdir(Path.joinpath(static_site_root, 'data'))
    os.mkdir(Path.joinpath(static_site_root, 'scripts'))

    with open(Path.joinpath(static_site_root, 'robots.txt'), 'w+') as f:
        f.writelines(['User-agent: *\n','Disallow: /'])
    
    # create index.html file
    reference_index_path = Path.joinpath(pathlib.Path(__file__).parent, '.bloggen/index.html')
    shutil.copy(reference_index_path,Path.joinpath(static_site_root, 'index.html'))

    print(site_info)
    f = open(Path.joinpath(static_site_root, 'data/site_info.json'), 'w+')
    json.dump(site_info, f)
    f.close()
    # with open(Path.joinpath(static_site_root, 'data/site_info.json'), 'w+')  as f:
    #     json.dump(site_info, f)


import os
from bloggen.config import Configure
from bloggen.static_site import Site 
from bloggen.site_info import Site_Info
import sys
import argparse
import pathlib

def main():
    config = Configure()
    parser = create_parser()
    if len(sys.argv) == 1:
        parser.print_help()
    else: 
        args = parser.parse_args()
        if args.config:
            argument: str = args.config
            if argument in ['no_args','create','new']:
                config.create_user_config()
            elif argument in ['list','ls']:
                config.list_config_names()
            elif argument.__contains__('='):
                key, value = argument.split('=')
                config.update_user_config(key,value)
            else:
                if config.user_config_exists(argument):
                    print(f"Activating {argument}")
                    config.set_active_config(argument)
                else:
                    print(f"User config {argument} does not exist.")
                    print(f"These configs exist:")
                    config.list_config_names()
        else:
            if not config.valid_user_config(config.active_config):
                pass
            else:
                site = Site(config)
                if args.add:
                    site.add(args.add)
                elif args.generate:
                    if args.generate == 'no_args':
                        site.generate()
                    else:
                        path_to_md_dir: str = args.generate
                        if path_to_md_dir[:3] == '../':
                            path_to_md_dir = clean_parent(path_to_md_dir)
                        site_info = Site_Info(path_to_md_dir, config.active_config["name"])
                        site.generate(pathlib.Path(path_to_md_dir), site_info.site_info)
                elif args.publish:
                    if args.publish == 'no_args':
                        site.publish()
                    else:
                        path_to_static_site = args.publish
                        if path_to_static_site[:3] == '../':
                            path_to_static_site = clean_parent(path_to_static_site)
                        site.publish(path_to_static_site)

def create_parser():
    parser = argparse.ArgumentParser(description="Create a static site!")
    parser.add_argument('-a', '--add', help="Upload a .md file.")
    parser.add_argument('--destroy', help="Terminates bucket!")
    parser.add_argument('-c','--config', help="Activate or create a configuration", nargs='?', const='no_args')
    parser.add_argument('-g','--generate', help="Builds static site locally. Provide path to md directory.", nargs='?', const='no_args')
    parser.add_argument('--remove', help="Removes a file from bucket.")
    parser.add_argument('--sync', help="Uploads directory to bucket.")
    parser.add_argument('-p','--publish', help="Uploads static site to bucket.", nargs='?', const='no_args')
    return parser

def clean_parent(path: str):
    count = path.count('../')
    cwd = pathlib.Path(os.getcwd())
    while count > 0:
       cwd = cwd.parent
       path = path[3:]
       count-=1
    return pathlib.Path.joinpath(cwd,path)

if __name__ == '__main__':
    main()



from datetime import datetime
from pathlib import Path
from typing import Dict, List
import uuid

directive_symbol_start = '<!--$'
directive_symbol_end = '$-->'
timestamp_format = '%m-%d-%YT%H:%M:%S'
class Site_Info:

    def __init__(self, target_dir, user:str):
        self.target_dir = target_dir
        self.user = user
        self.uuids = dict() # dict with keys: names values: uuids
        self.supported_directives = [{'name':'subjects', 'function': self.extract_subjects}, {'name':'tags', 'function': self.extract_tags}, {'name':'style', 'function': self.extract_style}]
        self.site_info = dict()
        self.get_site_info()

    def generate_ids(self, paths: List, id_tag: str, username: str, return_mapping: bool=False) -> List[str]:
        ids_paths: List[Dict[str,str]] = [{ 'id': self.generate_id(username, id_tag), 'path': path } for path in paths]
        if return_mapping:
            return [ele['id'] for ele in ids_paths], ids_paths
        return [ele['id'] for ele in ids_paths]

    def get_site_info(self) -> dict():
        relationship_graph, data, index = self.build_site_info(self.target_dir)
        self.site_info = { "relationship_graph": relationship_graph, "data": data, "index": index}

    def create_note(self, id, path:Path):
        directive_data = self.extract_data(path.absolute())
        return {
            'id': id,
            'content': path.as_uri(),
            'subjects': directive_data['subjects'],
            'tags':directive_data['tags'],
            'metadata': self.note_metadata()
        }

    def node_metadata(self):
        """
        As you decide to add more metadata objects, return more objects from here
        """
        return {'timestamp': datetime.now().strftime(timestamp_format)}
    
    def note_metadata(self):
        """
        As you decide to add more metadata objects, return more objects from here
        """
        style = 'default' # TODO extract style from the file. It is a directive. use extract_style()
        return {'timestamp': datetime.now().strftime(timestamp_format),'style':style}

    def build_site_info(self, dir_name) -> list[str]:

        def recurse_dirs(relationship_graph, data, index, blog_id: str, dir: Path):
            relationship_graph[blog_id] = {'blogs': [], 'notes': []}
            recursion_queue = []
            for f in dir.iterdir():
                if f.is_dir():
                    child_blog_id = self.generate_id(self.user, "blog")
                    index['blogs'].append(child_blog_id)
                    data['blogs'].append(child_blog_id)
                    relationship_graph[blog_id]['blogs'].append(child_blog_id)
                    recursion_queue.append((child_blog_id,f))
                elif f.name.endswith('.md'):
                    id = self.generate_id(self.user, "note")
                    path = f.absolute()
                    relationship_graph[blog_id]['notes'].append(id)
                    index['notes'].append(id)
                    data['notes'].append(self.create_note(id, path))
            for blog_id,dir in recursion_queue:
                relationship_graph, data, index = recurse_dirs(relationship_graph, data, index, blog_id, dir)

            return relationship_graph, data, index

        root_id = self.generate_id(self.user, "blog")
        path = Path(dir_name)
        return recurse_dirs({}, {"blogs": [], "notes": []}, {"blogs": [],"notes": []}, root_id, path)
    
    def generate_id(self, user, prefix: str) -> str:
        return f'{prefix}_{user}_{uuid.uuid4().hex[:5]}'

    def extract_data(self, filepath: str, ):
        with open(filepath, 'r') as f:
            contents = f.read()
        directive_data: Dict[str: [str]] = self.get_directives(contents, filepath)
        return directive_data
        # TODO add support for directives that are commands
        # return self.execute_commands(directive_data)

    def get_directives(self, contents: str, filepath: str) -> Dict[str, List[str]]:
        """
        At present, directive commands do not have context. They are simply pieces of text that are ignored by html
        """
        supported_directives = [s['name'] for s in self.supported_directives]
        directives = {directive:set() for directive in supported_directives}
        
        substr_start = 0

        directive_index = contents.find(directive_symbol_start, substr_start)
        while directive_index != -1:    
            colon_index = contents.find(':', directive_index)
            directive_end_index = contents.find(directive_symbol_end, colon_index)
            directive_start = directive_index + len(directive_symbol_start)
            directive: str = contents[directive_start:colon_index].strip(' ').lower()
            commands: List[str] = contents[colon_index+1:directive_end_index].split(',')
            commands = [command.strip(' ') for command in commands]
            # print(f'directive_command is {commands}')

            if directive not in supported_directives:
                print(f'{directive} is not a supported directive. Found {directive} in {filepath}')
            else:
                for command in commands:
                    directives.setdefault(directive, set()).add(command)
                    # else:
                    #     directives[directive] = set()
                    #     directives[directive].add(command)
                    
                    # print(f'directives: {directives}, command was {command}')
                    # directives.setdefault(directive, set()).add(command)

            substr_start = directive_end_index
            directive_index = contents.find(directive_symbol_start, substr_start)
        
        # we cannot write sets to json. This is a hacky solution to convert sets to lists
        for key in directives.keys():
                directives[key] = list(directives[key])
        print(f'directives: {directives}')
        return directives

    def execute_commands(self, found_directives: Dict[str, List[str]]):
        results = {}
        self.supported_directives
        for supported_directive in self.supported_directives: # [{'name':'subjects', 'function': self.extract_subjects}, {'name':'tags', 'function': self.extract_tags}]
            if supported_directive['name'] in found_directives:
                commands = found_directives[supported_directive['name']]
                for command in commands:
                    results[supported_directive['name']] = supported_directive['function'](command)
        return results

    def extract_tags(self, tag):
        return tag

    def extract_subjects(self,subject):
        return subject

    def extract_style(self, style):
        return style

    # def extract_data(self, nextraction_functions: Dict(str,List(Function))):
    #     # search for the symbol

    #     # goal: retrieve directives and their content

    #     #  return {{k, v(f)} for k,v in extraction_functions} # there is a method that yields the l and v of a single object here
    #     pass


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bloggen']

package_data = \
{'': ['*'], 'bloggen': ['.bloggen/*']}

install_requires = \
['Markdown>=3.4.1,<4.0.0',
 'argparse>=1.4,<2.0',
 'beautifulsoup4>=4.11.1,<5.0.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'google-cloud>=0.34.0,<0.35.0',
 'pathlib>=1.0.1,<2.0.0']

entry_points = \
{'console_scripts': ['bloggen = bloggen.__main__:main']}

setup_kwargs = {
    'name': 'bloggen',
    'version': '0.0.1',
    'description': 'Deploy a dir of md files as a static site',
    'long_description': "# Bloggen\n\nPoint bloggen to a hosting location, give it your markdown, then watch it deploy a static site.\n\n# Use\n\n## A parable:\n\nLee creates a GCP credentials json:\n\n1.  Lee gets a service token JSON from GCP. He follows this guide https://cloud.google.com/storage/docs/reference/libraries#setting_up_authentication'\n2.  Lee saves the JSON to his machine and remembers the path for later. Bloggen is going to save the path to that JSON into an environment variable\n3.  Lee will need to tell Bloggen the path to the json by configuring a Bloggen profile\n\nLee configures a profile:\n\n1. Lee runs `bloggen --config` and supplies the following information:\n\n   - Name\n   - Path to GCP credentials json\n   - Name of GCP bucket to use\n\n1. Lee writes 5 md notes into a dir\n1. Lee runs `bloggen --generate path_to_md_dir`\n1. Lee sees a new dir appear. It is named static-site and can be found at ../path_to_md_dir\n1. Lee runs `bloggen --publish path_to_static_site`\n1. Lee learns that his notes are available as a static site on GCP\n\n## Generate\n\n_generates static webpage_\nUser provides a directory of markdown to be converted to html\n\n`bloggen generate --path=path_to_dir`\n\n## Sync\n\n_Sync directory of markdown files with existing static site_\n`bloggen sync --path=path_to_dir_of_md_files`\n\n## Add\n\n_Add file to static site_\n_Convert given markdown file to html and add to site_\n`bloggen add --path=path_to_md_file`\n\n## Remove\n\n_Remove named page from static site_\n`bloggen remove --name=name_of_file`\n\n## Destroy\n\n_Destroy site and bucket_\n`bloggen destroy`\n",
    'author': 'Lee Harrold',
    'author_email': 'halzinnia@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
